/* Copyright (c) 2010 Michael Pippig
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "pfft.h"
#include "ipfft.h"

static void init_default_profile_non_transposed(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm);
static void init_default_profile_transposed(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm);
static void init_default_profile_like_p3dfft(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm);


void PX(init_many_input_3d)(
    const INT *n, const INT *ni, 
    const INT *local_n, const INT *local_start,
    MPI_Comm comm,
    C *data
    )
{
  int k0_inside, k1_inside, k2_inside, ind_inside, myrank;
  INT k0, k1, k2, i, j, k, m;
  INT zeros_size[3], zeros_start[3], zeros_end[3];
  R im, re;

  MPI_Comm_rank(comm, &myrank);
  srand(myrank);

  for(int t=0; t<3; t++)
    zeros_start[t] = (ni[t]+1)/2;
  PX(vsub_ptrdiff_t)(3, n, ni, zeros_size);
  PX(vadd_ptrdiff_t)(3, zeros_start, zeros_size, zeros_end);

  m=0;
  for(k0 = local_start[0]; k0 < local_start[0]+local_n[0]; k0++){
    k0_inside = (zeros_start[0] <= k0) && (k0 < zeros_end[0]);
    i = k0 - (zeros_end[0] <= k0) * zeros_size[0];
    for(k1 = local_start[1]; k1 < local_start[1]+local_n[1]; k1++){
      k1_inside = (zeros_start[1] <= k1) && (k1 < zeros_end[1]);
      j = k1 - (zeros_end[1] <= k1) * zeros_size[1];
      for(k2 = local_start[2]; k2 < local_start[2]+local_n[2]; k2++, m++){
        k2_inside = (zeros_start[2] <= k2) && (k2 < zeros_end[2]);
        k = k2 - (zeros_end[2] <= k2) * zeros_size[2];
        ind_inside = k0_inside || k1_inside || k2_inside;
        re = ind_inside ? 0 :  FFT_DATA_INIT;
        im = ind_inside ? 0 : -FFT_DATA_INIT;
        data[m][0] = re;
        data[m][1] = im;
      }
    }
  }
}


void PX(check_many_output_3d)(
    C *data, const INT *n, const INT *ni,
    const INT *local_ni, const INT *local_i_start,
    MPI_Comm comm, int iterations, int is_scaled
    )
{ /* check results after one forward and backward FFT, only works for intial data layout */
  INT i,j,k,m;
  R im, re;
  double error, maxerror, globmaxerr, scale;
  int myrank;

  MPI_Comm_rank(comm, &myrank);
  srand(myrank);

  /* data should be back in input layout */
  if( is_scaled == 0 )
    scale = 1.0/pow( n[0]*n[1]*n[2], iterations );
  else
    scale = 1;

  error = maxerror = 0;
  m=0;
  for(i = local_i_start[0]; i < local_i_start[0]+local_ni[0]; i++){
    for(j = local_i_start[1]; j < local_i_start[1]+local_ni[1]; j++){
      for(k = local_i_start[2]; k < local_i_start[2]+local_ni[2]; k++){
        re = scale*data[m][0]-FFT_DATA_INIT;
        im = scale*data[m][1]+FFT_DATA_INIT;
        error = sqrt(re*re + im*im);
        if( error > maxerror )
          maxerror = error;
        m++;
      } /* end for k */
    } /* end for j */
  } /* end for i */

  MPI_Allreduce(&maxerror, &globmaxerr, 1, MPI_DOUBLE, MPI_MAX, comm);
  if( myrank == 0 )
    fprintf(stderr, "maxerror = %6.2e;\n", globmaxerr);
}


INT PX(local_size_3d_many)(
    const INT *n, const INT *ni, const INT *no,
    INT howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    MPI_Comm comm_cart_2d, unsigned pfft_flags,
    INT *local_ni, INT *local_i_start,
    INT *local_no, INT *local_o_start,
    INT *alloc_local_data_in
    )
{ /* shortcut for forward & backward local_size_3d with non-transposed output */
  INT alloc_local_forw, alloc_local_back, alloc_local_gcells;
  PX(permutation) perm, perm_out;
  PX(profile) prf;

  /* memory needed for backward FFT */
  PX(init_default_profile)(pfft_flags| PFFT_BACKWARD, &prf, &perm);
  alloc_local_gcells = PX(local_size_3d_guru)(
      n, ni, no, &prf, &perm, howmany, block_col, block_row,
      num_gcells_below, num_gcells_above, comm_cart_2d,
      local_ni, local_i_start, local_no, local_o_start, &alloc_local_back, &perm_out);

  /* memory needed for forward FFT */
  PX(init_default_profile)(pfft_flags| PFFT_FORWARD, &prf, &perm);
  alloc_local_gcells = PX(local_size_3d_guru)(
      n, ni, no, &prf, &perm, howmany, block_col, block_row,
      num_gcells_below, num_gcells_above, comm_cart_2d,
      local_ni, local_i_start, local_no, local_o_start, &alloc_local_forw, &perm_out);

  *alloc_local_data_in = MAX(alloc_local_forw,alloc_local_back);
  return MAX(alloc_local_gcells,*alloc_local_data_in);
}


PX(plan) PX(plan_many_dft_3d)(
    const INT *n, const INT *ni, const INT *no,
    INT howmany, const INT *block_col, const INT *block_row,
    const INT *num_gcells_below, const INT *num_gcells_above,
    C *data_in, C *data_out,
    MPI_Comm comm_cart_2d,
    int sign, unsigned pfft_flags, unsigned fftw_flags
    )
{ /* !!! with flag FFTW_MEASURE planning will overwrite users data !!! */
  PX(permutation) perm;
  PX(profile) prf;
  
  PX(init_default_profile)(pfft_flags, &prf, &perm);
  
  return PX(plan_guru_dft_3d)(
      n, ni, no, &prf, &perm, howmany, block_col, block_row,
      num_gcells_below, num_gcells_above, data_in, data_out,
      comm_cart_2d, sign, fftw_flags);
}


int PX(check_many)(
    const INT *n, const INT *ni, const INT *no,
    const INT *block_col, const INT *block_row,
    MPI_Comm comm_cart_2d,
    unsigned pfft_flags, int verbose
    )
{
  int retval0, retval1;
  PX(permutation) perm;
  PX(profile) prf;

  /* check backward FFT */
  PX(init_default_profile)(pfft_flags| PFFT_BACKWARD, &prf, &perm);
  retval1 = PX(check_guru)(
      n, ni, no, &prf, &perm, block_col, block_row, comm_cart_2d, verbose);

  /* check forward FFT */
  PX(init_default_profile)(pfft_flags| PFFT_FORWARD, &prf, &perm);
  retval0 = PX(check_guru)(
      n, ni, no, &prf, &perm, block_col, block_row, comm_cart_2d, verbose);

  return MAX(retval0,retval1);
}


void PX(init_default_profile)(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm
    )
{
  if(pfft_flags & PFFT_TRANSPOSED)
    init_default_profile_transposed(pfft_flags, prf, perm);
  else if(pfft_flags & PFFT_P3DFFT)
    init_default_profile_like_p3dfft(pfft_flags, prf, perm);
  else
    init_default_profile_non_transposed(pfft_flags, prf, perm);
}


static void init_default_profile_non_transposed(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm
    )
{
  PX(profile_n012_np102)(prf);
  if(pfft_flags & PFFT_BACKWARD){
    PX(permute)(0, 1, 2, perm->n);
    PX(permute)(1, 0, 2, perm->np);
  }
  else{
    PX(permute)(0, 1, 2, perm->n);
    PX(permute)(0, 1, 2, perm->np);
  }
}


static void init_default_profile_transposed(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm
    )
{
/*  PX(profile_n021_np102_inplace)(prf);
  if(pfft_flags & PFFT_BACKWARD){
    PX(permute)(0, 2, 1, perm->n);
    PX(permute)(1, 0, 2, perm->np);
  }
  else{
    PX(permute)(0, 1, 2, perm->n);
    PX(permute)(0, 1, 2, perm->np);
  }*/
  
  PX(profile_n210_np102)(prf);
  if(pfft_flags & PFFT_BACKWARD){
    PX(permute)(2, 1, 0, perm->n);
    PX(permute)(1, 0, 2, perm->np);
  }
  else{
    PX(permute)(0, 1, 2, perm->n);
    PX(permute)(0, 1, 2, perm->np);
  }
}


static void init_default_profile_like_p3dfft(
    unsigned pfft_flags,
    PX(profile) *prf, PX(permutation) *perm
    )
{
  if(pfft_flags & PFFT_BACKWARD){
    PX(profile_n201_np012)(prf);
    PX(permute)(1, 2, 0, perm->n);
    PX(permute)(0, 1, 2, perm->np);
  }
  else{
    PX(profile_n120_np012_IN)(prf);
    PX(permute)(0, 1, 2, perm->n);
    PX(permute)(0, 1, 2, perm->np);
  }
}

