#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <math.h>
#include <complex.h>
/* return value 0 on success */

#include <fftw3.h>
#include <fftw3-mpi.h>
#include <pfft.h>

#define MAX(a,b) (((a)>(b))?(a):(b))

extern void get_arg_INT(
    int argc, char **argv, char *name, int neededArgs, ptrdiff_t *parameter);
extern void get_arg_int(
    int argc, char **argv, char *name, int neededArgs, int *parameter);

static int allocate_fftw_complex(
    ptrdiff_t alloc_local, fftw_complex **data, char *name);

static void pfft_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags, int verbose,
    double *time);
static void pfft_minimal_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags);


static void print_permuted_array_rank_ordered(
     double *data,
     ptrdiff_t local_n0, ptrdiff_t local_0_start,
     ptrdiff_t local_n1, ptrdiff_t local_1_start,
     ptrdiff_t local_n2, ptrdiff_t local_2_start,
     int perm0, int perm1, int perm2,
     char *name, int verbose,
     MPI_Comm comm);
static void print_permuted_array(
     double *data,
     ptrdiff_t local_n0, ptrdiff_t local_0_start,
     ptrdiff_t local_n1, ptrdiff_t local_1_start,
     ptrdiff_t local_n2, ptrdiff_t local_2_start,
     int perm0, int perm1, int perm2,
     char *name, int verbose);
static void print_array(
     double *data,
     ptrdiff_t *n, ptrdiff_t *start,
     char *name, int verbose);


static void init_parameters(
    int argc, char **argv,
    ptrdiff_t *n, int *np, ptrdiff_t *gcells_below, ptrdiff_t *gcells_above,
    int* planing, int* verbose
    )
{
  pfft_get_arg_INT(argc, argv, "-pfft_n", 3, n);
  pfft_get_arg_int(argc, argv, "-pfft_np", 2, np);
  pfft_get_arg_INT(argc, argv, "-pfft_gcells_below", 3, gcells_below);
  pfft_get_arg_INT(argc, argv, "-pfft_gcells_above", 3, gcells_above);
  pfft_get_arg_int(argc, argv, "-pfft_planing", 1, planing);
  pfft_get_arg_int(argc, argv, "-pfft_verbose", 1, verbose);
}


int main(int argc, char **argv){
  ptrdiff_t n[3], gcells_below[3], gcells_above[3];
  int np[2], myrank, size, planing, verbose;
  unsigned fftw_flags=0;
  double time_RMA[2];
  
  MPI_Init(&argc, &argv);
  pfft_init();
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  
  /* default values */
  n[0] = n[1] = n[2] = 128; /*  n[0] = 3; n[1] = 5; n[2] = 7;*/
  np[0]=size; np[1]=1;
  verbose = 0;
  for(int t=0; t<3; t++){
    gcells_below[t] = 4;
    gcells_above[t] = 5;
  }
  
  /* this option causes a bug with np=[2,2] and n=[3,3,3]*/
//   fftw_set_timelimit(120.0);
  
  /* set values by commandline */
  init_parameters(argc, argv, n, np, gcells_below, gcells_above, &planing, &verbose);

  switch(planing){
    case 0: fftw_flags = FFTW_ESTIMATE; break;
    case 2: fftw_flags = FFTW_PATIENT; break;
    case 3: fftw_flags = FFTW_EXHAUSTIVE; break;
    default: fftw_flags = FFTW_MEASURE;
  }

  pfft_test_gcells(
      n, np, gcells_below, gcells_above, MPI_COMM_WORLD, fftw_flags, verbose, time_RMA);
  pfft_minimal_test_gcells(
      n, np, gcells_below, gcells_above, MPI_COMM_WORLD, fftw_flags);

  if(!myrank)
    fprintf(stderr, "time_FFT_RMA = %.6e, time_RMA = %.6e\n",
        time_RMA[0], time_RMA[1]);
  
  pfft_cleanup();
  MPI_Finalize();
  return 0;
}



static void pfft_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags, int verbose,
    double *time
    )
{
  ptrdiff_t alloc_local_gc, alloc_local_data_in, local_ngc[3], local_gc_start[3];
  ptrdiff_t local_ni[3], local_i_start[3], local_no[3], local_o_start[3];
  fftw_complex *data;
  MPI_Comm comm_cart_2d;
  pfft_plan fftplan;
  ptrdiff_t howmany = 1;
  
  if( pfft_create_procmesh_2d(comm, np[0], np[1], &comm_cart_2d) )
    return;

  /* allocate mem for data */
  alloc_local_gc = pfft_local_size_3d_many(
      n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      gcells_below, gcells_above, comm_cart_2d, PFFT_TRANSPOSED,
      local_ni, local_i_start, local_no, local_o_start, &alloc_local_data_in);
  
  allocate_fftw_complex(alloc_local_gc, &data, "data");

  fftplan = pfft_plan_many_dft_3d(
    n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
    gcells_below, gcells_above, data, data, comm_cart_2d,
    FFTW_FORWARD, PFFT_TRANSPOSED| PFFT_FORWARD, fftw_flags);
  
  pfft_init_input_3d(n, local_ni, local_i_start, comm_cart_2d, data);
  time[0] = -MPI_Wtime();
  pfft_execute(fftplan);
  time[0] += MPI_Wtime();
  
  pfft_init_input_3d(n, local_no, local_o_start, comm_cart_2d, data);
  
  /* check gcell input */
  print_permuted_array_rank_ordered((double*) data, local_no[0], local_o_start[0], local_no[1], local_o_start[1],
      local_no[2], local_o_start[2], 2, 1, 0, "gcell input", verbose, comm);

  time[1] = -MPI_Wtime();
  /** add for gcell support */
  pfft_exchange_ghostcells(fftplan);
  time[1] += MPI_Wtime();
  
  for(int t=0; t<3; t++){
    local_ngc[t] = gcells_below[t] + local_no[t] + gcells_above[t];
    local_gc_start[t] = local_o_start[t] - gcells_below[t];
  }
  
  /* check output */
  print_permuted_array_rank_ordered((double*) data, local_ngc[0], local_gc_start[0], local_ngc[1], local_gc_start[1],
      local_ngc[2], local_gc_start[2], 2, 1, 0, "exchanged gcells", verbose, comm);
  
  pfft_reduce_ghostcells(fftplan);

  /* check input */
  print_permuted_array_rank_ordered((double*) data, local_no[0], local_o_start[0], local_no[1], local_o_start[1],
      local_no[2], local_o_start[2], 2, 1, 0, "reduced gcells", verbose, comm);
  
  /** free mem and finalize */
  pfft_destroy_plan(fftplan);
  fftw_free(data);
  MPI_Comm_free(&comm_cart_2d);
}


static void pfft_minimal_test_gcells(
    const ptrdiff_t *n, const int *np,
    const ptrdiff_t *gcells_below, const ptrdiff_t *gcells_above,
    MPI_Comm comm, unsigned fftw_flags
    )
{
  ptrdiff_t alloc_local_gc, alloc_local_data_in;
  ptrdiff_t local_ni[3], local_i_start[3], local_no[3], local_o_start[3];
  fftw_complex *data;
  MPI_Comm comm_cart_2d;
  pfft_plan plan_forw, plan_back;
  ptrdiff_t howmany = 1;

  if( pfft_create_procmesh_2d(comm, np[0], np[1], &comm_cart_2d) )
    return;

  alloc_local_gc = pfft_local_size_3d_many(
      n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
      gcells_below, gcells_above, comm_cart_2d, PFFT_TRANSPOSED,
      local_ni, local_i_start, local_no, local_o_start, &alloc_local_data_in);

  /* allocate mem for data */
  allocate_fftw_complex(alloc_local_gc, &data, "data");

  plan_forw = pfft_plan_many_dft_3d(
    n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
    gcells_below, gcells_above, data, data, comm_cart_2d,
    FFTW_FORWARD, PFFT_TRANSPOSED| PFFT_FORWARD, fftw_flags);
  plan_back = pfft_plan_many_dft_3d(
    n, n, n, howmany, PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS,
    PFFT_NO_GCELLS, PFFT_NO_GCELLS, data, data, comm_cart_2d,
    FFTW_BACKWARD, PFFT_TRANSPOSED| PFFT_BACKWARD, fftw_flags);

  /* initialize data */
  pfft_init_input_3d(n, local_ni, local_i_start, comm_cart_2d, data);

  /* compute parallel forward FFT */
  pfft_execute(plan_forw);

  /* send ghostcells */
  pfft_exchange_ghostcells(plan_forw);

  /* do some local calculations */

  /* sum up ghostcells */
  pfft_reduce_ghostcells(plan_forw);

  /* compute parallel backward FFT */
  pfft_execute(plan_back);

  /** free mem and finalize */
  pfft_destroy_plan(plan_forw);
  pfft_destroy_plan(plan_back);
  fftw_free(data);
  MPI_Comm_free(&comm_cart_2d);
}



static int allocate_fftw_complex(
    ptrdiff_t alloc_local, fftw_complex **data, char *name
    )
{
  int myrank, retval = 0;

  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

  *data = (fftw_complex*) fftw_malloc(sizeof(fftw_complex)*alloc_local);

  if( *data == NULL ){
    if(!myrank)
      fprintf(stderr, "!!! Error in PFFT: Not enough memory to allocate %s !!!\n", name);
    retval = 1;
  }

  return retval;
}


static void print_permuted_array_rank_ordered(
     double *data,
     ptrdiff_t local_n0, ptrdiff_t local_0_start,
     ptrdiff_t local_n1, ptrdiff_t local_1_start,
     ptrdiff_t local_n2, ptrdiff_t local_2_start,
     int perm0, int perm1, int perm2,
     char *name, int verbose,
     MPI_Comm comm
     )
{
  int num_procs, proc_rank;
  
  if(!verbose)
    return;
  
  MPI_Comm_size(comm, &num_procs);
  MPI_Comm_rank(comm, &proc_rank);

  fflush(stderr);
  for(int t=0; t<num_procs; t++){
    if(proc_rank == t){
      fprintf(stderr, "\nRank %d:\n", proc_rank);
      print_permuted_array(data, local_n0, local_0_start, local_n1, local_1_start,
          local_n2, local_2_start, perm0, perm1, perm2, name, verbose);
      fflush(stderr);
    }
    MPI_Barrier(comm);
  }
}


static void print_permuted_array(
     double *data,
     ptrdiff_t local_n0, ptrdiff_t local_0_start,
     ptrdiff_t local_n1, ptrdiff_t local_1_start,
     ptrdiff_t local_n2, ptrdiff_t local_2_start,
     int perm0, int perm1, int perm2,
     char *name, int verbose
     )
{
  ptrdiff_t local_start[3], local_n[3], perm[3];
  ptrdiff_t local_start_perm[3], local_n_perm[3];

  if(!verbose)
    return;
  
  local_n[0] = local_n0;  local_start[0] = local_0_start;
  local_n[1] = local_n1;  local_start[1] = local_1_start;
  local_n[2] = local_n2;  local_start[2] = local_2_start;
  
  perm[0] = perm0; perm[1] = perm1; perm[2] = perm2;
  
  for(int t=0; t<3; t++){
    local_n_perm[t] = local_n[perm[t]];
    local_start_perm[t] = local_start[perm[t]];
  }
  
  print_array(data, local_n_perm, local_start_perm, name, verbose);
}


static void print_array(
     double *data,
     ptrdiff_t *n, ptrdiff_t *start,
     char *name, int verbose
     )
{
  ptrdiff_t k0, k1, k2, h, l, howmany=2;
  
  if(!verbose)
    return;
  
  l=0;
  for(k0 = 0; k0 < n[0]; k0++){
    fprintf(stderr, "%s(%td,%td:%td,%td:%td):\n", name, k0+start[0]+1, start[1]+1, start[1]+n[1], start[2]+1, start[2]+n[2]);
    for(k1 = 0; k1 < n[1]; k1++){
      for(k2 = 0; k2 < n[2]; k2++){
        l = howmany*(k2 + n[2]*( k1 + n[1]*k0 ));
        fprintf(stderr, "%.1f", data[l++]);
        for(h=1; h<howmany; h++)
          fprintf(stderr, " , %.1f", data[l++]);
        fprintf(stderr, "\t");
      }
      fprintf(stderr, "\n");
    }
  }
}

